@@ -44,7 +44,7 @@ struct _EphyAboutHandler {
   EphySMaps *smaps;
 };
 
-G_DEFINE_TYPE (EphyAboutHandler, ephy_about_handler, G_TYPE_OBJECT)
+G_DEFINE_FINAL_TYPE (EphyAboutHandler, ephy_about_handler, G_TYPE_OBJECT)
 
 
 #define EPHY_ABOUT_OVERVIEW_MAX_ITEMS 9
@@ -164,14 +164,26 @@ ephy_about_handler_handle_about (EphyAboutHandler       *handler,
 {
   char *data;
   char *version;
-  GtkIconInfo *icon_info;
+  g_autofree char *path = NULL;
+  GtkIconTheme *icon_theme;
+  g_autoptr (GtkIconPaintable) paintable = NULL;
 
   version = g_strdup_printf (_("Version %s"), VERSION);
 
-  icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
+  icon_theme = gtk_icon_theme_get_for_display (gdk_display_get_default ());
+  paintable = gtk_icon_theme_lookup_icon (icon_theme,
                                           APPLICATION_ID,
+                                          NULL,
                                           256,
-                                          GTK_ICON_LOOKUP_FORCE_SVG);
+                                          1,
+                                          GTK_TEXT_DIR_LTR,
+                                          GTK_ICON_LOOKUP_FORCE_REGULAR);
+
+  if (paintable) {
+    g_autoptr (GFile) file = gtk_icon_paintable_get_file (paintable);
+
+    path = g_file_get_path (file);
+  }
 
   data = g_strdup_printf ("<html><head><title>%s</title>"
                           "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />"
@@ -188,7 +200,7 @@ ephy_about_handler_handle_about (EphyAboutHandler       *handler,
                           "</table>"
                           "</div></div></body></html>",
                           _("About Web"),
-                          icon_info ? gtk_icon_info_get_filename (icon_info) : "",
+                          path ? path : "",
 #if !TECH_PREVIEW
                           _("Web"),
 #else
@@ -198,8 +210,6 @@ ephy_about_handler_handle_about (EphyAboutHandler       *handler,
                           _("A simple, clean, beautiful view of the web"),
                           "WebKitGTK", webkit_get_major_version (), webkit_get_minor_version (), webkit_get_micro_version ());
   g_free (version);
-  if (icon_info)
-    g_object_unref (icon_info);
 
   ephy_about_handler_finish_request (request, data, -1);
 
@@ -256,61 +266,88 @@ handle_applications_finished_cb (EphyAboutHandler       *handler,
                             "</script>"
                             "</head><div id=\"applications\"><body class=\"applications-body\"><h1>%s</h1>"
                             "<p>%s</p>",
-                            _("Applications"),
-                            _("Applications"),
-                            _("List of installed web applications"));
+                            _("Apps"),
+                            _("Apps"),
+                            _("List of installed web apps"));
 
     g_string_append (data_str, "<table>");
 
     for (p = applications; p; p = p->next) {
       EphyWebApplication *app = (EphyWebApplication *)p->data;
-      g_autofree char *encoded_icon_url = NULL;
+      const char *icon_path = NULL;
+      g_autofree char *encoded_icon_path = NULL;
       g_autofree char *encoded_name = NULL;
       g_autofree char *encoded_url = NULL;
-      g_autofree char *js_encoded_id = NULL;
+      g_autoptr (GDate) date = NULL;
+      char install_date[128];
 
       if (ephy_web_application_is_system (app))
         continue;
 
-      /* Most of these fields are at least semi-trusted. The web app suggests
-       * its own title, which gets used in the app ID and icon URL, but it ought
-       * to be safe because we validate that it is a valid component of a
-       * GApplication ID, which should not permit anything nasty. The icon URL
-       * could be changed by the user to something else after web app creation,
-       * though, so better not fully trust it. Then the app name and the main
-       * URL could contain contain anything at all, so those need to be encoded
-       * for sure. Install date should be fine because it's constructed by
-       * Epiphany.
+      date = g_date_new ();
+      g_date_set_time_t (date, (time_t)app->install_date_uint64);
+      g_date_strftime (install_date, 127, "%x", date);
+
+      /* In the sandbox we don't have access to the host side icon file */
+      if (ephy_is_running_inside_sandbox ())
+        icon_path = app->tmp_icon_path;
+      else
+        icon_path = app->icon_path;
+
+      if (!icon_path) {
+        g_warning ("Failed to get icon path for app %s", app->id);
+        continue;
+      }
+
+      /* Most of these fields are at least semi-trusted. The app ID was chosen
+       * by ephy so it's safe. The icon URL could be changed by the user to
+       * something else after web app creation, though, so better not fully
+       * trust it. Then the app name and the main URL could contain contain
+       * anything at all, so those need to be encoded for sure. Install date
+       * should be fine because it's constructed by Epiphany.
        */
-      encoded_icon_url = ephy_encode_for_html_attribute (app->icon_url);
+      encoded_icon_path = ephy_encode_for_html_attribute (icon_path);
       encoded_name = ephy_encode_for_html_entity (app->name);
       encoded_url = ephy_encode_for_html_entity (app->url);
       g_string_append_printf (data_str,
                               "<tbody><tr id =\"%s\">"
                               "<td class=\"icon\"><img width=64 height=64 src=\"file://%s\"></img></td>"
                               "<td class=\"data\"><div class=\"appname\">%s</div><div class=\"appurl\">%s</div></td>"
-                              "<td class=\"input\"><input type=\"button\" value=\"%s\" onclick=\"deleteWebApp('%s');\"></td>"
+                              "<td class=\"input\"><input type=\"button\" value=\"%s\" onclick=\"deleteWebApp('%s');\" "
+                              "class=\"destructive-action\"></td>"
                               "<td class=\"date\">%s <br /> %s</td></tr></tbody>",
-                              app->id, encoded_icon_url, encoded_name, encoded_url, _("Delete"), app->id,
+                              app->id, encoded_icon_path, encoded_name, encoded_url, _("Delete"), app->id,
                               /* Note for translators: this refers to the installation date. */
-                              _("Installed on:"), app->install_date);
+                              _("Installed on:"), install_date);
     }
 
     g_string_append (data_str, "</table></div></body></html>");
   } else {
-    g_autoptr (GtkIconInfo) icon_info = NULL;
-    g_autofree gchar *icon = g_strconcat ("application-x-addon-symbolic", NULL);
+    GtkIconTheme *icon_theme;
+    g_autoptr (GtkIconPaintable) paintable = NULL;
+    g_autofree char *path = NULL;
 
     g_string_append_printf (data_str, "<html><head><title>%s</title>"
                             "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />"
                             "<link href=\""EPHY_PAGE_TEMPLATE_ABOUT_CSS "\" rel=\"stylesheet\" type=\"text/css\">"
                             "</head><body class=\"applications-body\">",
-                            _("Applications"));
+                            _("Apps"));
 
-    icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
-                                            icon,
+    icon_theme = gtk_icon_theme_get_for_display (gdk_display_get_default ());
+    paintable = gtk_icon_theme_lookup_icon (icon_theme,
+                                            "application-x-addon-symbolic",
+                                            NULL,
                                             128,
+                                            1,
+                                            GTK_TEXT_DIR_LTR,
                                             0);
+
+    if (paintable) {
+      g_autoptr (GFile) file = gtk_icon_paintable_get_file (paintable);
+
+      path = g_file_get_path (file);
+    }
+
     g_string_append_printf (data_str,
                             "  <div id=\"overview\" class=\"overview-empty\">\n"
                             "    <img src=\"file://%s\"/>\n"
@@ -318,9 +355,9 @@ handle_applications_finished_cb (EphyAboutHandler       *handler,
                             "    <div><p>%s</p></div>\n"
                             "  </div>\n"
                             "</body></html>\n",
-                            icon_info ? gtk_icon_info_get_filename (icon_info) : "",
+                            path ? path : "",
                             /* Displayed when opening applications without any installed web apps. */
-                            _("Applications"), _("You can add your favorite website by clicking <b>Install Site as Web Application…</b> within the page menu."));
+                            _("Apps"), _("You can add your favorite website by clicking <b>Install as Web App…</b> within the page menu."));
   }
 
   ephy_web_application_free_application_list (applications);
@@ -390,19 +427,32 @@ history_service_query_urls_cb (EphyHistoryService     *history,
                           "<body>\n",
                           lang, lang,
                           ((gtk_widget_get_default_direction () == GTK_TEXT_DIR_RTL) ? "rtl" : "ltr"),
-                          _(OVERVIEW_PAGE_TITLE));
+                          _(NEW_TAB_PAGE_TITLE));
   g_free (lang);
 
   list_length = g_list_length (urls);
 
   if (list_length == 0 || !success) {
-    GtkIconInfo *icon_info;
-    g_autofree gchar *icon = g_strconcat (APPLICATION_ID, "-symbolic", NULL);
+    GtkIconTheme *icon_theme;
+    g_autoptr (GtkIconPaintable) paintable = NULL;
+    g_autofree char *path = NULL;
+    g_autofree char *icon = g_strconcat (APPLICATION_ID, "-symbolic", NULL);
 
-    icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
+    icon_theme = gtk_icon_theme_get_for_display (gdk_display_get_default ());
+    paintable = gtk_icon_theme_lookup_icon (icon_theme,
                                             icon,
+                                            NULL,
                                             128,
+                                            1,
+                                            GTK_TEXT_DIR_LTR,
                                             0);
+
+    if (paintable) {
+      g_autoptr (GFile) file = gtk_icon_paintable_get_file (paintable);
+
+      path = g_file_get_path (file);
+    }
+
     g_string_append_printf (data_str,
                             "  <div id=\"overview\" class=\"overview-empty\">\n"
                             "    <img src=\"file://%s\"/>\n"
@@ -410,11 +460,9 @@ history_service_query_urls_cb (EphyHistoryService     *history,
                             "    <div><p>%s</p></div>\n"
                             "  </div>\n"
                             "</body></html>\n",
-                            icon_info ? gtk_icon_info_get_filename (icon_info) : "",
+                            path ? path : "",
                             /* Displayed when opening the browser for the first time. */
                             _("Welcome to Web"), _("Start browsing and your most-visited sites will appear here."));
-    if (icon_info)
-      g_object_unref (icon_info);
     goto out;
   }
 
@@ -464,6 +512,23 @@ out:
   g_object_unref (request);
 }
 
+static gboolean
+ephy_about_handler_handle_newtab (EphyAboutHandler       *handler,
+                                  WebKitURISchemeRequest *request)
+{
+  char *data;
+
+  data = g_strdup_printf ("<html><head><title>%s</title>"
+                          "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />"
+                          "</head><body style=\"color-scheme: light dark;\">"
+                          "</body></html>",
+                          _(NEW_TAB_PAGE_TITLE));
+
+  ephy_about_handler_finish_request (request, data, -1);
+
+  return TRUE;
+}
+
 EphyHistoryQuery *
 ephy_history_query_new_for_overview (void)
 {
@@ -565,8 +630,10 @@ ephy_about_handler_handle_request (EphyAboutHandler       *handler,
     handled = ephy_about_handler_handle_memory (handler, request);
   else if (!g_strcmp0 (path, "epiphany"))
     handled = ephy_about_handler_handle_epiphany (handler, request);
-  else if (!g_strcmp0 (path, "applications") && !ephy_is_running_inside_flatpak ())
+  else if (!g_strcmp0 (path, "applications"))
     handled = ephy_about_handler_handle_applications (handler, request);
+  else if (!g_strcmp0 (path, "newtab"))
+    handled = ephy_about_handler_handle_newtab (handler, request);
   else if (!g_strcmp0 (path, "overview"))
     handled = ephy_about_handler_handle_html_overview (handler, request);
   else if (!g_strcmp0 (path, "incognito"))
