@@ -25,7 +25,6 @@
 #include "ephy-embed.h"
 #include "ephy-embed-shell.h"
 #include "ephy-embed-type-builtins.h"
-#include "ephy-file-chooser.h"
 #include "ephy-file-helpers.h"
 #include "ephy-flatpak-utils.h"
 #include "ephy-prefs.h"
@@ -40,26 +39,38 @@ struct _EphyDownload {
   GObject parent_instance;
 
   WebKitDownload *download;
+  GCancellable *cancellable;
 
   char *content_type;
+  char *suggested_directory;
+  char *suggested_filename;
 
   gboolean show_notification;
+  gboolean always_ask_destination;
+  gboolean choose_filename;
 
   EphyDownloadActionType action;
-  guint32 start_time;
   gboolean finished;
   GError *error;
   GFileMonitor *file_monitor;
+
+  guint64 uid;
+
+  char *initiated_by_extension_id;
+  char *initiated_by_extension_name;
+
+  GDateTime *start_time;
+  GDateTime *end_time;
+  gboolean was_moved;
 };
 
-G_DEFINE_TYPE (EphyDownload, ephy_download, G_TYPE_OBJECT)
+G_DEFINE_FINAL_TYPE (EphyDownload, ephy_download, G_TYPE_OBJECT)
 
 enum {
   PROP_0,
   PROP_DOWNLOAD,
   PROP_DESTINATION,
   PROP_ACTION,
-  PROP_START_TIME,
   PROP_CONTENT_TYPE,
   LAST_PROP
 };
@@ -76,6 +87,8 @@ enum {
 
 static guint signals[LAST_SIGNAL];
 
+static guint64 download_uid = 1;
+
 static void
 ephy_download_get_property (GObject    *object,
                             guint       property_id,
@@ -89,14 +102,11 @@ ephy_download_get_property (GObject    *object,
       g_value_set_object (value, ephy_download_get_webkit_download (download));
       break;
     case PROP_DESTINATION:
-      g_value_set_string (value, ephy_download_get_destination_uri (download));
+      g_value_set_string (value, ephy_download_get_destination (download));
       break;
     case PROP_ACTION:
       g_value_set_enum (value, ephy_download_get_action (download));
       break;
-    case PROP_START_TIME:
-      g_value_set_uint (value, ephy_download_get_start_time (download));
-      break;
     case PROP_CONTENT_TYPE:
       g_value_set_string (value, ephy_download_get_content_type (download));
       break;
@@ -117,13 +127,12 @@ ephy_download_set_property (GObject      *object,
 
   switch (property_id) {
     case PROP_DESTINATION:
-      ephy_download_set_destination_uri (download, g_value_get_string (value));
+      ephy_download_set_destination (download, g_value_get_string (value));
       break;
     case PROP_ACTION:
       ephy_download_set_action (download, g_value_get_enum (value));
       break;
     case PROP_DOWNLOAD:
-    case PROP_START_TIME:
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -198,20 +207,23 @@ parse_extension (const char *filename)
 }
 
 static gboolean
-set_destination_uri_for_suggested_filename (EphyDownload *download,
-                                            const char   *suggested_filename)
+set_destination_for_suggested_filename (EphyDownload *download,
+                                        const char   *directory,
+                                        const char   *suggested_filename)
 {
   char *dest_dir;
   char *dest_name;
-  char *destination_filename;
-  char *destination_uri;
+  g_autofree char *destination_filename = NULL;
 
-  dest_dir = ephy_file_get_downloads_dir ();
+  if (directory)
+    dest_dir = g_strdup (directory);
+  else
+    dest_dir = ephy_file_get_downloads_dir ();
 
   /* Make sure the download directory exists */
   if (g_mkdir_with_parents (dest_dir, 0700) == -1) {
-    g_critical ("Could not create downloads directory \"%s\": %s",
-                dest_dir, strerror (errno));
+    g_warning ("Could not create downloads directory \"%s\": %s",
+               dest_dir, strerror (errno));
     g_free (dest_dir);
     return FALSE;
   }
@@ -227,7 +239,7 @@ set_destination_uri_for_suggested_filename (EphyDownload *download,
   g_free (dest_name);
 
   /* Append (n) as needed. */
-  if (g_file_test (destination_filename, G_FILE_TEST_EXISTS)) {
+  if (!webkit_download_get_allow_overwrite (download->download) && g_file_test (destination_filename, G_FILE_TEST_EXISTS)) {
     int i = 1;
     const char *dot_pos;
     gssize position;
@@ -255,30 +267,24 @@ set_destination_uri_for_suggested_filename (EphyDownload *download,
     g_string_free (tmp_filename, TRUE);
   }
 
-  destination_uri = g_filename_to_uri (destination_filename, NULL, NULL);
-  g_free (destination_filename);
-
-  g_assert (destination_uri);
-  webkit_download_set_destination (download->download, destination_uri);
-  g_free (destination_uri);
+  webkit_download_set_destination (download->download, destination_filename);
 
   return TRUE;
 }
 
 /**
- * ephy_download_set_destination_uri:
+ * ephy_download_set_destination:
  * @download: an #EphyDownload
- * @destination: URI where to save @download
+ * @destination: path at which to save @download
  *
- * Sets the destination URI of @download. It must be a proper URI, with a
- * scheme like file:/// or similar.
+ * Sets the destination of @download.
  **/
 void
-ephy_download_set_destination_uri (EphyDownload *download,
-                                   const char   *destination)
+ephy_download_set_destination (EphyDownload *download,
+                               const char   *destination)
 {
   g_assert (EPHY_IS_DOWNLOAD (download));
-  g_assert (destination != NULL);
+  g_assert (destination);
 
   webkit_download_set_destination (download->download, destination);
   g_object_notify_by_pspec (G_OBJECT (download), obj_properties[PROP_DESTINATION]);
@@ -320,15 +326,15 @@ ephy_download_get_webkit_download (EphyDownload *download)
 }
 
 /**
- * ephy_download_get_destination_uri:
+ * ephy_download_get_destination:
  * @download: an #EphyDownload
  *
- * Gets the destination URI where the download is being saved.
+ * Gets the destination where the download is being saved.
  *
- * Returns: (transfer none): destination URI.
+ * Returns: (transfer none): destination path
  **/
 const char *
-ephy_download_get_destination_uri (EphyDownload *download)
+ephy_download_get_destination (EphyDownload *download)
 {
   g_assert (EPHY_IS_DOWNLOAD (download));
 
@@ -354,23 +360,6 @@ ephy_download_get_action (EphyDownload *download)
   return download->action;
 }
 
-/**
- * ephy_download_get_start_time:
- * @download: an #EphyDownload
- *
- * Gets the time (returned by gtk_get_current_event_time ()) when @download was
- * started. Defaults to 0.
- *
- * Returns: the time when @download was started.
- **/
-guint32
-ephy_download_get_start_time (EphyDownload *download)
-{
-  g_assert (EPHY_IS_DOWNLOAD (download));
-
-  return download->start_time;
-}
-
 /**
  * ephy_download_cancel:
  * @download: an #EphyDownload
@@ -420,7 +409,6 @@ ephy_download_failed (EphyDownload  *download,
  * ephy_download_do_download_action:
  * @download: an #EphyDownload
  * @action: one of #EphyDownloadActionType
- * @user_time: GDK timestamp, for focus-stealing prevention
  *
  * Executes the given @action for @download, this can be any of
  * #EphyDownloadActionType.
@@ -430,27 +418,26 @@ ephy_download_failed (EphyDownload  *download,
  **/
 gboolean
 ephy_download_do_download_action (EphyDownload           *download,
-                                  EphyDownloadActionType  action,
-                                  guint32                 user_time)
+                                  EphyDownloadActionType  action)
 {
   GFile *destination;
-  const char *destination_uri;
+  const char *destination_path;
   gboolean ret = FALSE;
 
-  destination_uri = webkit_download_get_destination (download->download);
-  destination = g_file_new_for_uri (destination_uri);
+  destination_path = webkit_download_get_destination (download->download);
+  destination = g_file_new_for_path (destination_path);
 
   switch ((action ? action : download->action)) {
     case EPHY_DOWNLOAD_ACTION_BROWSE_TO:
       LOG ("ephy_download_do_download_action: browse_to");
       /* Must not use this action type under sandbox! */
-      ret = ephy_file_browse_to (destination, user_time);
+      ret = ephy_file_browse_to (destination);
       break;
     case EPHY_DOWNLOAD_ACTION_OPEN:
       LOG ("ephy_download_do_download_action: open");
-      ret = ephy_file_launch_handler (destination, user_time);
+      ret = ephy_file_launch_handler (destination);
       if (!ret)
-        ret = ephy_file_browse_to (destination, user_time);
+        ret = ephy_file_browse_to (destination);
       break;
     case EPHY_DOWNLOAD_ACTION_NONE:
       LOG ("ephy_download_do_download_action: none");
@@ -477,9 +464,18 @@ ephy_download_dispose (GObject *object)
     download->download = NULL;
   }
 
+  g_cancellable_cancel (download->cancellable);
+  g_clear_object (&download->cancellable);
+
   g_clear_object (&download->file_monitor);
   g_clear_error (&download->error);
   g_clear_pointer (&download->content_type, g_free);
+  g_clear_pointer (&download->suggested_filename, g_free);
+  g_clear_pointer (&download->suggested_directory, g_free);
+  g_clear_pointer (&download->start_time, g_date_time_unref);
+  g_clear_pointer (&download->end_time, g_date_time_unref);
+  g_clear_pointer (&download->initiated_by_extension_id, g_free);
+  g_clear_pointer (&download->initiated_by_extension_name, g_free);
 
   G_OBJECT_CLASS (ephy_download_parent_class)->dispose (object);
 }
@@ -500,8 +496,7 @@ ephy_download_class_init (EphyDownloadClass *klass)
    */
   obj_properties[PROP_DOWNLOAD] =
     g_param_spec_object ("download",
-                         "Internal WebKitDownload",
-                         "The WebKitDownload used internally by EphyDownload",
+                         NULL, NULL,
                          WEBKIT_TYPE_DOWNLOAD,
                          G_PARAM_READABLE |
                          G_PARAM_STATIC_STRINGS);
@@ -513,8 +508,7 @@ ephy_download_class_init (EphyDownloadClass *klass)
    */
   obj_properties[PROP_DESTINATION] =
     g_param_spec_string ("destination",
-                         "Destination",
-                         "Destination file URI",
+                         NULL, NULL,
                          NULL,
                          G_PARAM_READWRITE |
                          G_PARAM_STATIC_STRINGS);
@@ -527,31 +521,15 @@ ephy_download_class_init (EphyDownloadClass *klass)
    */
   obj_properties[PROP_ACTION] =
     g_param_spec_enum ("action",
-                       "Download action",
-                       "Action to take when download finishes",
+                       NULL, NULL,
                        EPHY_TYPE_DOWNLOAD_ACTION_TYPE,
                        EPHY_DOWNLOAD_ACTION_NONE,
                        G_PARAM_READABLE |
                        G_PARAM_STATIC_STRINGS);
 
-  /**
-   * EphyDownload::start-time:
-   *
-   * User time when the download started, useful for launching applications
-   * aware of focus stealing.
-   */
-  obj_properties[PROP_START_TIME] =
-    g_param_spec_uint ("start-time",
-                       "Event start time",
-                       "Time for focus-stealing prevention.",
-                       0, G_MAXUINT32, 0,
-                       G_PARAM_READABLE |
-                       G_PARAM_STATIC_STRINGS);
-
   obj_properties[PROP_CONTENT_TYPE] =
     g_param_spec_string ("content-type",
-                         "Content Type",
-                         "The download content type",
+                         NULL, NULL,
                          NULL,
                          G_PARAM_READABLE |
                          G_PARAM_STATIC_STRINGS);
@@ -562,14 +540,18 @@ ephy_download_class_init (EphyDownloadClass *klass)
    * EphyDownload::filename-suggested:
    *
    * The ::filename-suggested signal is emitted when we have received the
-   * suggested filename from WebKit.
+   * suggested filename from WebKit. Return %TRUE if you will provide a
+   * destination or will %FALSE otherwise. If the destination is not
+   * provided before the signal handler returns, the download will not
+   * start until provided.
    **/
   signals[FILENAME_SUGGESTED] = g_signal_new ("filename-suggested",
                                               G_OBJECT_CLASS_TYPE (object_class),
                                               G_SIGNAL_RUN_LAST,
                                               0,
-                                              NULL, NULL, NULL,
-                                              G_TYPE_NONE,
+                                              g_signal_accumulator_true_handled,
+                                              NULL, NULL,
+                                              G_TYPE_BOOLEAN,
                                               1,
                                               G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);
 
@@ -618,12 +600,252 @@ ephy_download_init (EphyDownload *download)
   LOG ("EphyDownload initialising %p", download);
 
   download->download = NULL;
+  download->cancellable = g_cancellable_new ();
 
   download->action = EPHY_DOWNLOAD_ACTION_NONE;
 
-  download->start_time = gtk_get_current_event_time ();
-
   download->show_notification = TRUE;
+
+  download->uid = download_uid++;
+}
+
+typedef struct {
+  EphyDownload *download;
+  char *suggested_filename;
+  GtkWindow *dialog;
+  GFile *directory;
+  GtkLabel *directory_label;
+  gboolean choose_filename;
+} SuggestedFilenameData;
+
+static SuggestedFilenameData *
+suggested_filename_data_new (EphyDownload *download,
+                             const char   *suggested_filename,
+                             GtkWindow    *dialog,
+                             GFile        *directory,
+                             GtkLabel     *directory_label,
+                             gboolean      choose_filename)
+{
+  SuggestedFilenameData *data = g_new (SuggestedFilenameData, 1);
+
+  data->download = g_object_ref (download);
+  data->suggested_filename = g_strdup (suggested_filename);
+  data->dialog = dialog;
+  data->directory = g_object_ref (directory);
+  data->directory_label = directory_label;
+  data->choose_filename = choose_filename;
+
+  return data;
+}
+
+static void
+suggested_filename_data_free (SuggestedFilenameData *data)
+{
+  g_object_unref (data->download);
+  g_object_unref (data->directory);
+  g_free (data->suggested_filename);
+  g_free (data);
+}
+
+static void
+filename_suggested_dialog_cb (AdwMessageDialog      *dialog,
+                              const char            *response,
+                              SuggestedFilenameData *data)
+{
+  if (!strcmp (response, "download")) {
+    g_autofree gchar *directory = g_file_get_path (data->directory);
+    WebKitDownload *webkit_download = ephy_download_get_webkit_download (data->download);
+
+    set_destination_for_suggested_filename (data->download, directory, data->suggested_filename);
+
+    webkit_download_set_allow_overwrite (webkit_download, TRUE);
+
+    ephy_downloads_manager_add_download (ephy_embed_shell_get_downloads_manager (ephy_embed_shell_get_default ()),
+                                         data->download);
+
+    g_settings_set_string (EPHY_SETTINGS_WEB, EPHY_PREFS_WEB_LAST_DOWNLOAD_DIRECTORY, directory);
+  } else {
+    ephy_download_cancel (data->download);
+  }
+
+  suggested_filename_data_free (data);
+}
+
+static void
+filename_suggested_file_dialog_cb (GtkFileDialog         *dialog,
+                                   GAsyncResult          *result,
+                                   SuggestedFilenameData *data)
+{
+  g_autoptr (GFile) file = NULL;
+  g_autofree char *display_name = NULL;
+  g_autoptr (GError) error = NULL;
+
+  if (!data->choose_filename)
+    file = gtk_file_dialog_select_folder_finish (dialog, result, &error);
+  else
+    file = gtk_file_dialog_save_finish (dialog, result, &error);
+
+  if (!file) {
+    g_warning ("Failed to select download destination: %s", error->message);
+    return;
+  }
+
+  g_set_object (&data->directory, file);
+
+  display_name = ephy_file_get_display_name (data->directory);
+  gtk_label_set_label (data->directory_label, display_name);
+}
+
+static void
+filename_suggested_button_cb (GtkButton             *button,
+                              SuggestedFilenameData *data)
+{
+  GtkFileDialog *dialog = gtk_file_dialog_new ();
+
+  gtk_file_dialog_set_initial_folder (dialog, data->directory);
+
+  if (!data->choose_filename) {
+    gtk_file_dialog_set_title (dialog, _("Select a Directory"));
+
+    gtk_file_dialog_select_folder (dialog,
+                                   data->dialog,
+                                   data->download->cancellable,
+                                   (GAsyncReadyCallback)filename_suggested_file_dialog_cb,
+                                   data);
+  } else {
+    gtk_file_dialog_set_title (dialog, _("Select the Destination"));
+    gtk_file_dialog_set_initial_name (dialog, data->suggested_filename);
+
+    gtk_file_dialog_save (dialog,
+                          data->dialog,
+                          data->download->cancellable,
+                          (GAsyncReadyCallback)filename_suggested_file_dialog_cb,
+                          data);
+  }
+}
+
+static void
+open_download_confirmation_dialog (EphyDownload *download,
+                                   const char   *suggested_filename)
+{
+  GApplication *application;
+  GtkWidget *dialog = NULL;
+  GtkWidget *grid;
+  GtkWindow *toplevel;
+  GtkWidget *icon;
+  GtkWidget *name_label;
+  GtkWidget *type_label;
+  GtkWidget *source_label;
+  GtkWidget *question_label;
+  GtkWidget *button;
+  GtkWidget *button_box;
+  GtkWidget *button_label;
+  WebKitDownload *webkit_download;
+  WebKitURIResponse *response;
+  g_autofree gchar *type_text = NULL;
+  g_autofree gchar *source_text = NULL;
+  g_autofree gchar *content_length = NULL;
+  g_autofree gchar *display_name = NULL;
+  g_autoptr (GIcon) gicon = NULL;
+  g_autoptr (GFile) directory = NULL;
+  const gchar *content_type;
+  const char *directory_path;
+  SuggestedFilenameData *data;
+
+  application = G_APPLICATION (ephy_embed_shell_get_default ());
+  toplevel = gtk_application_get_active_window (GTK_APPLICATION (application));
+
+  dialog = adw_message_dialog_new (GTK_WINDOW (toplevel),
+                                   _("Download Requested"),
+                                   NULL);
+  adw_message_dialog_add_responses (ADW_MESSAGE_DIALOG (dialog),
+                                    "cancel", _("_Cancel"),
+                                    "download", _("_Download"),
+                                    NULL);
+
+  webkit_download = ephy_download_get_webkit_download (download);
+  response = webkit_download_get_response (webkit_download);
+
+  grid = gtk_grid_new ();
+  gtk_grid_set_column_spacing (GTK_GRID (grid), 6);
+  gtk_widget_set_margin_top (grid, 6);
+  adw_message_dialog_set_extra_child (ADW_MESSAGE_DIALOG (dialog), grid);
+
+  content_length = g_format_size (webkit_uri_response_get_content_length (response));
+  content_type = ephy_download_get_content_type (download);
+
+  /* Icon */
+  gicon = g_content_type_get_symbolic_icon (content_type);
+  icon = gtk_image_new_from_gicon (gicon);
+  gtk_image_set_pixel_size (GTK_IMAGE (icon), 64);
+  gtk_grid_attach (GTK_GRID (grid), icon, 0, 0, 1, 3);
+
+  /* Name */
+  name_label = gtk_label_new (suggested_filename);
+  gtk_label_set_ellipsize (GTK_LABEL (name_label), PANGO_ELLIPSIZE_END);
+  gtk_label_set_xalign (GTK_LABEL (name_label), 0);
+  gtk_widget_add_css_class (name_label, "heading");
+  gtk_grid_attach (GTK_GRID (grid), name_label, 1, 0, 1, 1);
+
+  /* Type */
+  type_text = g_strdup_printf (_("Type: %s (%s)"), g_content_type_get_description (content_type), content_length);
+  type_label = gtk_label_new (type_text);
+  gtk_label_set_xalign (GTK_LABEL (type_label), 0);
+  gtk_grid_attach (GTK_GRID (grid), type_label, 1, 1, 1, 1);
+
+  /* Source */
+  source_text = g_strdup_printf (_("Source: %s"), ephy_string_get_host_name (webkit_uri_response_get_uri (response)));
+  source_label = gtk_label_new (source_text);
+  gtk_label_set_xalign (GTK_LABEL (source_label), 0);
+  gtk_grid_attach (GTK_GRID (grid), source_label, 1, 2, 1, 1);
+
+  /* Question */
+  question_label = gtk_label_new (_("Where do you want to save the file?"));
+  gtk_widget_set_margin_top (question_label, 18);
+  gtk_widget_set_margin_bottom (question_label, 6);
+  gtk_grid_attach (GTK_GRID (grid), question_label, 0, 3, 2, 1);
+
+  /* File Chooser Button */
+  button = gtk_button_new ();
+  gtk_grid_attach (GTK_GRID (grid), button, 0, 4, 2, 1);
+
+  button_box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
+  gtk_widget_set_hexpand (button_box, FALSE);
+  gtk_button_set_child (GTK_BUTTON (button), button_box);
+
+  gtk_box_append (GTK_BOX (button_box),
+                  gtk_image_new_from_icon_name ("folder-symbolic"));
+
+  button_label = gtk_label_new (NULL);
+  gtk_label_set_ellipsize (GTK_LABEL (button_label), PANGO_ELLIPSIZE_END);
+  gtk_label_set_xalign (GTK_LABEL (button_label), 0);
+  gtk_widget_set_hexpand (button_label, TRUE);
+  gtk_box_append (GTK_BOX (button_box), button_label);
+
+  directory_path = g_settings_get_string (EPHY_SETTINGS_WEB, EPHY_PREFS_WEB_LAST_DOWNLOAD_DIRECTORY);
+  if (download->suggested_directory)
+    directory = g_file_new_for_path (download->suggested_directory);
+  else if (!directory_path || !directory_path[0])
+    directory = g_file_new_for_path (ephy_file_get_downloads_dir ());
+  else
+    directory = g_file_new_for_path (directory_path);
+
+  display_name = ephy_file_get_display_name (directory);
+  gtk_label_set_label (GTK_LABEL (button_label), display_name);
+
+  data = suggested_filename_data_new (download,
+                                      suggested_filename,
+                                      GTK_WINDOW (dialog),
+                                      directory,
+                                      GTK_LABEL (button_label),
+                                      download->choose_filename);
+
+  g_signal_connect (button, "clicked",
+                    G_CALLBACK (filename_suggested_button_cb), data);
+  g_signal_connect (dialog, "response",
+                    G_CALLBACK (filename_suggested_dialog_cb), data);
+
+  gtk_window_present (GTK_WINDOW (dialog));
 }
 
 static void
@@ -646,18 +868,31 @@ download_response_changed_cb (WebKitDownload *wk_download,
 
 static gboolean
 download_decide_destination_cb (WebKitDownload *wk_download,
-                                const gchar    *suggested_filename,
+                                const gchar    *wk_suggestion,
                                 EphyDownload   *download)
 {
-  if (webkit_download_get_destination (wk_download))
-    return TRUE;
+  const char *suggested_filename = wk_suggestion;
+  gboolean will_provide_destination = FALSE;
 
-  g_signal_emit (download, signals[FILENAME_SUGGESTED], 0, suggested_filename);
+  if (download->suggested_filename)
+    suggested_filename = download->suggested_filename;
 
   if (webkit_download_get_destination (wk_download))
     return TRUE;
 
-  return set_destination_uri_for_suggested_filename (download, suggested_filename);
+  g_signal_emit (download, signals[FILENAME_SUGGESTED], 0, suggested_filename, &will_provide_destination);
+
+  if (will_provide_destination)
+    return TRUE;
+
+  if (!ephy_is_running_inside_sandbox () &&
+      (g_settings_get_boolean (EPHY_SETTINGS_WEB, EPHY_PREFS_WEB_ASK_ON_DOWNLOAD) ||
+       download->always_ask_destination)) {
+    open_download_confirmation_dialog (download, suggested_filename);
+    return TRUE;
+  }
+
+  return set_destination_for_suggested_filename (download, download->suggested_directory, suggested_filename);
 }
 
 static void
@@ -668,6 +903,8 @@ download_created_destination_cb (WebKitDownload *wk_download,
   char *filename;
   char *content_type;
 
+  download->start_time = g_date_time_new_now_local ();
+
   if (download->content_type && !g_content_type_is_unknown (download->content_type))
     return;
 
@@ -742,9 +979,11 @@ download_file_monitor_changed (GFileMonitor      *monitor,
                                EphyDownload      *download)
 {
   /* Skip messages for <file>.wkdownload */
-  if (strcmp (g_file_get_uri (file), webkit_download_get_destination (download->download)) != 0)
+  if (strcmp (g_file_peek_path (file), webkit_download_get_destination (download->download)) != 0)
     return;
 
+  download->was_moved = TRUE;
+
   if (event_type == G_FILE_MONITOR_EVENT_DELETED || event_type == G_FILE_MONITOR_EVENT_MOVED)
     g_signal_emit (download, signals[MOVED], 0);
 }
@@ -757,15 +996,16 @@ download_finished_cb (WebKitDownload *wk_download,
   g_autoptr (GFile) file = NULL;
 
   download->finished = TRUE;
+  download->end_time = g_date_time_new_now_local ();
 
-  ephy_download_do_download_action (download, download->action, download->start_time);
+  ephy_download_do_download_action (download, download->action);
 
   if (download->show_notification)
     display_download_finished_notification (wk_download);
 
   g_signal_emit (download, signals[COMPLETED], 0);
 
-  file = g_file_new_for_uri (webkit_download_get_destination (wk_download));
+  file = g_file_new_for_path (webkit_download_get_destination (wk_download));
   download->file_monitor = g_file_monitor (file, G_FILE_MONITOR_NONE, NULL, &error);
   if (!download->file_monitor)
     g_warning ("Could not add a file monitor for %s, error: %s\n", g_file_get_uri (file), error->message);
@@ -782,95 +1022,11 @@ download_failed_cb (WebKitDownload *wk_download,
 
   LOG ("error (%d - %d)! %s", error->code, 0, error->message);
   download->finished = TRUE;
+  download->end_time = g_date_time_new_now_local ();
   download->error = g_error_copy (error);
   g_signal_emit (download, signals[ERROR], 0, download->error);
 }
 
-static void
-filename_suggested_cb (EphyDownload *download,
-                       const char   *suggested_filename,
-                       gpointer      user_data)
-{
-  GApplication *application;
-  GtkWidget *dialog = NULL;
-  GtkWidget *message_area;
-  GtkWidget *box;
-  GtkWindow *toplevel;
-  GtkWidget *type_label;
-  GtkWidget *from_label;
-  GtkWidget *question_label;
-  GtkWidget *filechooser;
-  WebKitDownload *webkit_download;
-  WebKitURIResponse *response;
-  g_autofree gchar *type_text = NULL;
-  g_autofree gchar *from_text = NULL;
-  g_autofree gchar *content_length = NULL;
-  const gchar *content_type;
-
-  application = G_APPLICATION (ephy_embed_shell_get_default ());
-  toplevel = gtk_application_get_active_window (GTK_APPLICATION (application));
-
-  dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
-                                   GTK_DIALOG_USE_HEADER_BAR | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
-                                   GTK_MESSAGE_QUESTION,
-                                   GTK_BUTTONS_NONE,
-                                   "%s",
-                                   _("Download requested"));
-  gtk_dialog_add_buttons (GTK_DIALOG (dialog), _("_Cancel"), GTK_RESPONSE_CANCEL, _("_Download"), GTK_RESPONSE_OK, NULL);
-  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), "%s", suggested_filename);
-  message_area = gtk_message_dialog_get_message_area (GTK_MESSAGE_DIALOG (dialog));
-
-  webkit_download = ephy_download_get_webkit_download (download);
-  response = webkit_download_get_response (webkit_download);
-
-  box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
-  gtk_box_pack_start (GTK_BOX (message_area), box, FALSE, TRUE, 0);
-
-  /* Type */
-  content_length = g_format_size (webkit_uri_response_get_content_length (response));
-  content_type = ephy_download_get_content_type (download);
-  type_text = g_strdup_printf (_("Type: %s (%s)"), g_content_type_get_description (content_type), content_length);
-  type_label = gtk_label_new (type_text);
-  gtk_widget_set_margin_top (type_label, 12);
-  gtk_box_pack_start (GTK_BOX (box), type_label, FALSE, TRUE, 0);
-
-  /* From */
-  from_text = g_strdup_printf (_("From: %s"), ephy_string_get_host_name (webkit_uri_response_get_uri (response)));
-  from_label = gtk_label_new (from_text);
-  gtk_box_pack_start (GTK_BOX (box), from_label, FALSE, TRUE, 0);
-
-  /* Question */
-  question_label = gtk_label_new (_("Where do you want to save the file?"));
-  gtk_widget_set_margin_top (question_label, 12);
-  gtk_box_pack_start (GTK_BOX (box), question_label, FALSE, TRUE, 0);
-
-  /* File Chooser Button */
-  filechooser = gtk_file_chooser_button_new (_("Save file"), GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
-  gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (filechooser), g_settings_get_string (EPHY_SETTINGS_WEB, EPHY_PREFS_WEB_LAST_DOWNLOAD_DIRECTORY));
-  gtk_box_pack_start (GTK_BOX (box), filechooser, FALSE, TRUE, 0);
-
-  gtk_widget_show_all (box);
-
-  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK) {
-    g_autofree gchar *uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (filechooser));
-    g_autofree gchar *folder = g_filename_from_uri (uri, NULL, NULL);
-    g_autofree gchar *path = g_build_filename (uri, suggested_filename, NULL);
-
-    ephy_download_set_destination_uri (download, path);
-
-    webkit_download_set_allow_overwrite (webkit_download, TRUE);
-
-    ephy_downloads_manager_add_download (ephy_embed_shell_get_downloads_manager (ephy_embed_shell_get_default ()),
-                                         download);
-
-    g_settings_set_string (EPHY_SETTINGS_WEB, EPHY_PREFS_WEB_LAST_DOWNLOAD_DIRECTORY, folder);
-  } else {
-    ephy_download_cancel (download);
-  }
-
-  gtk_widget_destroy (dialog);
-}
-
 EphyDownload *
 ephy_download_new_internal (WebKitDownload *download)
 {
@@ -918,12 +1074,6 @@ ephy_download_new (WebKitDownload *download)
                            G_CALLBACK (download_decide_destination_cb),
                            ephy_download, 0);
 
-  if (!ephy_is_running_inside_sandbox () && g_settings_get_boolean (EPHY_SETTINGS_WEB, EPHY_PREFS_WEB_ASK_ON_DOWNLOAD)) {
-    g_signal_connect (ephy_download, "filename-suggested",
-                      G_CALLBACK (filename_suggested_cb),
-                      NULL);
-  }
-
   return ephy_download;
 }
 
@@ -944,7 +1094,7 @@ ephy_download_new_for_uri (const char *uri)
 
   g_assert (uri != NULL);
 
-  download = webkit_web_context_download_uri (ephy_embed_shell_get_web_context (shell), uri);
+  download = webkit_network_session_download_uri (ephy_embed_shell_get_network_session (shell), uri);
   ephy_download = ephy_download_new (download);
   g_object_unref (download);
 
@@ -960,7 +1110,7 @@ ephy_download_new_for_uri_internal (const char *uri)
 
   g_assert (uri != NULL);
 
-  download = webkit_web_context_download_uri (ephy_embed_shell_get_web_context (shell), uri);
+  download = webkit_network_session_download_uri (ephy_embed_shell_get_network_session (shell), uri);
   ephy_download = ephy_download_new_internal (download);
 
   return ephy_download;
@@ -973,3 +1123,178 @@ ephy_download_disable_desktop_notification (EphyDownload *download)
 
   download->show_notification = FALSE;
 }
+
+guint64
+ephy_download_get_uid (EphyDownload *download)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  return download->uid;
+}
+
+/**
+ * ephy_download_set_always_ask_destination:
+ *
+ * Bypasses the global user preference for prompting for
+ * a save location. This does not bypass EphyDownload:destination
+ * being set.
+ */
+void
+ephy_download_set_always_ask_destination (EphyDownload *download,
+                                          gboolean      always_ask)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  download->always_ask_destination = always_ask;
+}
+
+/**
+ * ephy_download_set_choose_filename:
+ *
+ * Changes the download prompt to select the destination
+ * filename rather than only the directory.
+ */
+void
+ephy_download_set_choose_filename (EphyDownload *download,
+                                   gboolean      choose_filename)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  download->choose_filename = choose_filename;
+}
+
+/**
+ * ephy_download_set_suggested_destination:
+ * @suggested_directory: (nullable): Default download directory
+ * @suggested_filename: (nullable): Default filename
+ *
+ * This sets recommendations for the directory and filename of
+ * the download. If the directory does not exist it will be created.
+ * The filename will be sanitized and possibly renamed if needed.
+ *
+ * If @suggested_directory is %NULL the globally configured download
+ * directory is used.
+ *
+ * If @suggested_filename is %NULL the WebKit recommended filename
+ * is used.
+ *
+ * Note that this does not override EphyDownload:destination and only
+ * provides default suggestions.
+ */
+void
+ephy_download_set_suggested_destination (EphyDownload *download,
+                                         const char   *suggested_directory,
+                                         const char   *suggested_filename)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  g_free (download->suggested_directory);
+  download->suggested_directory = g_strdup (suggested_directory);
+
+  g_free (download->suggested_filename);
+  download->suggested_filename = g_strdup (suggested_filename);
+}
+
+/**
+ * ephy_download_set_allow_overwrite:
+ *
+ * This allows the downloaded file to overwrite files on disk and
+ * also disables the automatic renaming (appending "(1)") when the file
+ * already exists.
+ */
+void
+ephy_download_set_allow_overwrite (EphyDownload *download,
+                                   gboolean      allow_overwrite)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  webkit_download_set_allow_overwrite (download->download, allow_overwrite);
+}
+
+/**
+ * ephy_download_get_was_moved:
+ *
+ * Returns: %TRUE if Epiphany detected the file being moved or deleted
+ */
+gboolean
+ephy_download_get_was_moved (EphyDownload *download)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  return download->was_moved;
+}
+
+/**
+ * ephy_download_get_start_time:
+ *
+ * Returns: (nullable): The time the download was started or %NULL
+ */
+GDateTime *
+ephy_download_get_start_time (EphyDownload *download)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  return download->start_time;
+}
+
+/**
+ * ephy_download_get_end_time:
+ *
+ * Returns: (nullable): The time the download was completed/failed or %NULL if active
+ */
+GDateTime *
+ephy_download_get_end_time (EphyDownload *download)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  return download->end_time;
+}
+
+/**
+ * ephy_download_get_initiating_web_extension_info:
+ * @download: The #EphyDownload
+ * @extension_id_out: (nullable): Place to store the extension ID
+ * @extension_name_out: (nullable): Place to store the extension name
+ *
+ * This returns information on which, if any, WebExtension created the download.
+ *
+ * Returns: %TRUE if web extension info exists
+ */
+gboolean
+ephy_download_get_initiating_web_extension_info (EphyDownload  *download,
+                                                 const char   **extension_id_out,
+                                                 const char   **extension_name_out)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  if (extension_name_out)
+    *extension_name_out = download->initiated_by_extension_name ? download->initiated_by_extension_name : NULL;
+
+  if (extension_id_out)
+    *extension_id_out = download->initiated_by_extension_id ? download->initiated_by_extension_id : NULL;
+
+  return download->initiated_by_extension_name || download->initiated_by_extension_id;
+}
+
+/**
+ * ephy_download_set_initiating_web_extension_info:
+ * @download: The #EphyDownload
+ * @extension_id: (nullable): Extension ID
+ * @extension_name: (nullable): Extension name
+ *
+ * This sets that @download was created by a WebExtension. This information is exposed to
+ * other WebExtensions.
+ */
+void
+ephy_download_set_initiating_web_extension_info (EphyDownload *download,
+                                                 const char   *extension_id,
+                                                 const char   *extension_name)
+{
+  g_assert (EPHY_IS_DOWNLOAD (download));
+
+  g_free (download->initiated_by_extension_name);
+  download->initiated_by_extension_name = g_strdup (extension_name);
+
+  g_free (download->initiated_by_extension_id);
+  download->initiated_by_extension_id = g_strdup (extension_id);
+}
