@@ -89,7 +89,7 @@ struct _EphyEmbed {
   gboolean progress_bar_enabled;
 };
 
-G_DEFINE_TYPE (EphyEmbed, ephy_embed, GTK_TYPE_BOX)
+G_DEFINE_FINAL_TYPE (EphyEmbed, ephy_embed, GTK_TYPE_BOX)
 
 enum {
   PROP_0,
@@ -380,8 +380,16 @@ static void
 ephy_embed_finalize (GObject *object)
 {
   EphyEmbed *embed = EPHY_EMBED (object);
+  EphyEmbedShell *shell = ephy_embed_shell_get_default ();
   GSList *list;
 
+  g_signal_handlers_disconnect_by_func (shell, ephy_embed_restored_window_cb, embed);
+
+  list = embed->destroy_on_transition_list;
+  for (; list; list = list->next) {
+    GtkWidget *widget = GTK_WIDGET (list->data);
+    g_signal_handlers_disconnect_by_func (widget, remove_from_destroy_list_cb, embed);
+  }
   g_slist_free (embed->destroy_on_transition_list);
 
   for (list = embed->messages; list; list = list->next) {
@@ -619,7 +627,7 @@ progress_update (EphyWebView *view,
                                  (loading || progress == 1.0) ? progress : 0.0);
 }
 
-static gboolean
+static void
 load_delayed_request_if_mapped (gpointer user_data)
 {
   EphyEmbed *embed = EPHY_EMBED (user_data);
@@ -629,7 +637,7 @@ load_delayed_request_if_mapped (gpointer user_data)
   embed->delayed_request_source_id = 0;
 
   if (!gtk_widget_get_mapped (GTK_WIDGET (embed)))
-    return G_SOURCE_REMOVE;
+    return;
 
   web_view = ephy_embed_get_web_view (embed);
   if (embed->delayed_state)
@@ -644,7 +652,12 @@ load_delayed_request_if_mapped (gpointer user_data)
   g_clear_object (&embed->delayed_request);
   g_clear_pointer (&embed->delayed_state, webkit_web_view_session_state_unref);
 
-  return G_SOURCE_REMOVE;
+  /* We have a binding to `is-loading` in `ephy_tab_view_add_tab ()` that depends on
+   * whether the page is a placeholder (to avoid showing the spinner on tabs while restoring the session),
+   * so after removing the placeholder we need to notify it again. */
+  g_object_notify (G_OBJECT (web_view), "is-loading");
+
+  return;
 }
 
 static void
@@ -657,7 +670,7 @@ ephy_embed_maybe_load_delayed_request (EphyEmbed *embed)
    * is scrolling rapidly through a bunch of delayed tabs, we don't start
    * loading them all.
    */
-  embed->delayed_request_source_id = g_timeout_add (300, load_delayed_request_if_mapped, embed);
+  embed->delayed_request_source_id = g_timeout_add_once (300, load_delayed_request_if_mapped, embed);
   g_source_set_name_by_id (embed->delayed_request_source_id, "[epiphany] load_delayed_request_if_mapped");
 }
 
@@ -717,8 +730,8 @@ ephy_embed_constructed (GObject *object)
   WebKitWebInspector *inspector;
   GtkEventController *controller;
 
-  g_signal_connect_object (shell, "window-restored",
-                           G_CALLBACK (ephy_embed_restored_window_cb), embed, 0);
+  g_signal_connect (shell, "window-restored",
+                    G_CALLBACK (ephy_embed_restored_window_cb), embed);
 
   g_signal_connect (embed, "map",
                     G_CALLBACK (ephy_embed_mapped_cb), NULL);
@@ -756,41 +769,39 @@ ephy_embed_constructed (GObject *object)
   }
 
   embed->find_toolbar = ephy_find_toolbar_new (embed->web_view);
-  g_signal_connect_object (embed->find_toolbar, "close",
-                           G_CALLBACK (ephy_embed_find_toolbar_close_cb),
-                           embed, 0);
+  g_signal_connect (embed->find_toolbar, "close",
+                    G_CALLBACK (ephy_embed_find_toolbar_close_cb),
+                    embed);
 
   gtk_box_append (GTK_BOX (embed), GTK_WIDGET (embed->find_toolbar));
 
   if (embed->progress_bar_enabled)
-    embed->progress_update_handler_id = g_signal_connect_object (embed->web_view, "notify::estimated-load-progress",
-                                                                 G_CALLBACK (progress_update), embed, 0);
+    embed->progress_update_handler_id = g_signal_connect (embed->web_view, "notify::estimated-load-progress",
+                                                          G_CALLBACK (progress_update), object);
 
   gtk_box_append (GTK_BOX (embed), GTK_WIDGET (embed->top_widgets_vbox));
   gtk_box_append (GTK_BOX (embed), GTK_WIDGET (embed->overlay));
 
-  g_signal_connect_object (embed->web_view, "notify::title",
-                           G_CALLBACK (web_view_title_changed_cb), embed, 0);
-  g_signal_connect_object (embed->web_view, "load-changed",
-                           G_CALLBACK (load_changed_cb), embed, 0);
-  g_signal_connect_object (embed->web_view, "enter-fullscreen",
-                           G_CALLBACK (entering_fullscreen_cb), embed, 0);
-  g_signal_connect_object (embed->web_view, "leave-fullscreen",
-                           G_CALLBACK (leaving_fullscreen_cb), embed, 0);
+  g_object_connect (embed->web_view,
+                    "signal::notify::title", G_CALLBACK (web_view_title_changed_cb), embed,
+                    "signal::load-changed", G_CALLBACK (load_changed_cb), embed,
+                    "signal::enter-fullscreen", G_CALLBACK (entering_fullscreen_cb), embed,
+                    "signal::leave-fullscreen", G_CALLBACK (leaving_fullscreen_cb), embed,
+                    NULL);
 
-  embed->status_handler_id = g_signal_connect_object (embed->web_view, "notify::status-message",
-                                                      G_CALLBACK (status_message_notify_cb),
-                                                      embed, 0);
+  embed->status_handler_id = g_signal_connect (embed->web_view, "notify::status-message",
+                                               G_CALLBACK (status_message_notify_cb),
+                                               embed);
 
   /* The inspector */
   inspector = webkit_web_view_get_inspector (embed->web_view);
 
-  g_signal_connect_object (inspector, "attach",
-                           G_CALLBACK (ephy_embed_attach_inspector_cb),
-                           embed, 0);
-  g_signal_connect_object (inspector, "closed",
-                           G_CALLBACK (ephy_embed_close_inspector_cb),
-                           embed, 0);
+  g_signal_connect (inspector, "attach",
+                    G_CALLBACK (ephy_embed_attach_inspector_cb),
+                    embed);
+  g_signal_connect (inspector, "closed",
+                    G_CALLBACK (ephy_embed_close_inspector_cb),
+                    embed);
 
   if (webkit_web_view_is_controlled_by_automation (embed->web_view)) {
     GtkWidget *banner;
@@ -874,7 +885,7 @@ ephy_embed_add_top_widget (EphyEmbed                *embed,
     list = g_slist_prepend (list, widget);
     embed->destroy_on_transition_list = list;
 
-    g_signal_connect_object (widget, "destroy", G_CALLBACK (remove_from_destroy_list_cb), embed, 0);
+    g_signal_connect (widget, "destroy", G_CALLBACK (remove_from_destroy_list_cb), embed);
   }
 
   gtk_box_prepend (embed->top_widgets_vbox, widget);
