@@ -22,6 +22,7 @@
 
 #include "ephy-bookmark-properties.h"
 #include "ephy-bookmark-row.h"
+#include "ephy-bookmarks-popover.h"
 #include "ephy-embed-prefs.h"
 #include "ephy-embed-shell.h"
 #include "ephy-favicon-helpers.h"
@@ -38,7 +39,7 @@ struct _EphyBookmarkRow {
   GtkWidget *properties_button;
 };
 
-G_DEFINE_TYPE (EphyBookmarkRow, ephy_bookmark_row, GTK_TYPE_LIST_BOX_ROW)
+G_DEFINE_FINAL_TYPE (EphyBookmarkRow, ephy_bookmark_row, GTK_TYPE_LIST_BOX_ROW)
 
 enum {
   PROP_0,
@@ -55,16 +56,16 @@ ephy_bookmark_row_button_clicked_cb (EphyBookmarkRow *row,
   GtkWidget *popover;
   GtkWidget *dialog;
   GtkWidget *grid;
+  GtkWidget *toolbar_view;
   GtkEventController *controller;
   GtkShortcut *shortcut;
 
   g_assert (EPHY_IS_BOOKMARK_ROW (row));
   g_assert (GTK_IS_BUTTON (button));
 
-  dialog = g_object_new (GTK_TYPE_WINDOW,
+  dialog = g_object_new (ADW_TYPE_WINDOW,
                          "title", _("Bookmark Properties"),
                          "transient-for", GTK_WINDOW (gtk_widget_get_root (GTK_WIDGET (row))),
-                         "titlebar", gtk_header_bar_new (),
                          "modal", TRUE,
                          NULL);
 
@@ -87,11 +88,25 @@ ephy_bookmark_row_button_clicked_cb (EphyBookmarkRow *row,
   gtk_window_set_default_widget (GTK_WINDOW (dialog),
                                  ephy_bookmark_properties_get_add_tag_button (EPHY_BOOKMARK_PROPERTIES (grid)));
 
-  gtk_window_set_child (GTK_WINDOW (dialog), grid);
+  toolbar_view = adw_toolbar_view_new ();
+  adw_toolbar_view_add_top_bar (ADW_TOOLBAR_VIEW (toolbar_view),
+                                adw_header_bar_new ());
+  adw_toolbar_view_set_content (ADW_TOOLBAR_VIEW (toolbar_view),
+                                grid);
+  adw_window_set_content (ADW_WINDOW (dialog), toolbar_view);
 
   gtk_window_present (GTK_WINDOW (dialog));
 }
 
+void
+ephy_bookmark_row_open (EphyBookmarkRow *self,
+                        EphyLinkFlags    flags)
+{
+  GtkWidget *window = gtk_widget_get_ancestor (GTK_WIDGET (self), EPHY_TYPE_WINDOW);
+  const char *url = ephy_bookmark_get_url (self->bookmark);
+  ephy_link_open (EPHY_LINK (window), url, NULL, flags | EPHY_LINK_BOOKMARK);
+}
+
 static void
 ephy_bookmark_row_favicon_loaded_cb (GObject      *source,
                                      GAsyncResult *result,
@@ -190,12 +205,27 @@ transform_bookmark_title (GBinding     *binding,
   return TRUE;
 }
 
+static void
+ephy_bookmark_row_map (GtkWidget *widget)
+{
+  EphyBookmarkRow *self = EPHY_BOOKMARK_ROW (widget);
+  EphyEmbedShell *shell = ephy_embed_shell_get_default ();
+  WebKitFaviconDatabase *database;
+
+  GTK_WIDGET_CLASS (ephy_bookmark_row_parent_class)->map (widget);
+
+  database = ephy_embed_shell_get_favicon_database (shell);
+  webkit_favicon_database_get_favicon (database,
+                                       ephy_bookmark_get_url (self->bookmark),
+                                       self->cancellable,
+                                       (GAsyncReadyCallback)ephy_bookmark_row_favicon_loaded_cb,
+                                       self);
+}
+
 static void
 ephy_bookmark_row_constructed (GObject *object)
 {
   EphyBookmarkRow *self = EPHY_BOOKMARK_ROW (object);
-  EphyEmbedShell *shell = ephy_embed_shell_get_default ();
-  WebKitFaviconDatabase *database;
 
   G_OBJECT_CLASS (ephy_bookmark_row_parent_class)->constructed (object);
 
@@ -211,13 +241,16 @@ ephy_bookmark_row_constructed (GObject *object)
                    self->properties_button,
                    "visible",
                    G_SETTINGS_BIND_INVERT_BOOLEAN);
+}
 
-  database = ephy_embed_shell_get_favicon_database (shell);
-  webkit_favicon_database_get_favicon (database,
-                                       ephy_bookmark_get_url (self->bookmark),
-                                       self->cancellable,
-                                       (GAsyncReadyCallback)ephy_bookmark_row_favicon_loaded_cb,
-                                       self);
+static void
+ephy_bookmark_row_activate (GtkListBoxRow *row)
+{
+  GtkWidget *popover = gtk_widget_get_ancestor (GTK_WIDGET (row), EPHY_TYPE_BOOKMARKS_POPOVER);
+  if (popover) {
+    gtk_popover_popdown (GTK_POPOVER (popover));
+  }
+  ephy_bookmark_row_open (EPHY_BOOKMARK_ROW (row), 0);
 }
 
 static void
@@ -231,6 +264,10 @@ ephy_bookmark_row_class_init (EphyBookmarkRowClass *klass)
   object_class->dispose = ephy_bookmark_row_dispose;
   object_class->constructed = ephy_bookmark_row_constructed;
 
+  widget_class->map = ephy_bookmark_row_map;
+
+  GTK_LIST_BOX_ROW_CLASS (klass)->activate = ephy_bookmark_row_activate;
+
   obj_properties[PROP_BOOKMARK] =
     g_param_spec_object ("bookmark",
                          NULL, NULL,
